#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

#include <windows.h>

#define TIMER_INTERVAL 200
#define TIMER_ID 1

static void error_dialog(const char *message) {
	char text[strlen(message) + 2 + 11 + 2];
	sprintf(text, "%s: %d.", message, (int) GetLastError());
	MessageBoxA(NULL, text, NULL, MB_ICONERROR);
}

struct icon_t {
	HDC dc;
	HFONT font;
	HBITMAP bitmap;
	HBITMAP mask;
	bool caps;
};

static HICON create_icon(struct icon_t *icon, LCID lcid) {
	char name[9] = "Zz";
	if (lcid && GetLocaleInfoA(lcid, LOCALE_SISO639LANGNAME,
		name, sizeof(name))) {
		name[2] = '\0';
		for (int i = 0; name[i]; i++) {
			name[i] = (i == 0 || icon->caps
				? toupper : tolower)(name[i]);
		}
	}
	BITMAP bitmap = {
		.bmWidth = 0,
		.bmHeight = 0
	};
	if (!GetObject(icon->bitmap, sizeof(bitmap), &bitmap)) {
		return NULL;
	}
	int width = bitmap.bmWidth;
	int height = bitmap.bmHeight;
	SelectObject(icon->dc, icon->bitmap);
	FillRect(icon->dc, &(RECT) { 0, 0, width, height },
		(HBRUSH) (COLOR_ACTIVECAPTION + 1));
	SelectObject(icon->dc, icon->font);
	SetTextColor(icon->dc, GetSysColor(COLOR_CAPTIONTEXT));
	DrawTextA(icon->dc, name, strlen(name),
		&(RECT) { 0, 0, width, height },
		DT_SINGLELINE | DT_CENTER | DT_VCENTER);
	return CreateIconIndirect(&(ICONINFO) {
		.fIcon = TRUE,
		.hbmMask = icon->mask,
		.hbmColor = icon->bitmap
	});
}

enum exit_t {
	EXIT_PASS,
	EXIT_DISABLE,
	EXIT_CONFIRM
};

struct state_t {
	HWND focus;
	LCID lcid;
	struct icon_t icon;
	NOTIFYICONDATAA notify;
	bool notify_add;
	enum exit_t exit;
};

static void update_state(struct state_t *state, HWND window) {
	HWND focus = GetForegroundWindow();
	if (focus != state->focus) {
		if (focus == window) {
			focus = NULL;
		} else {
			static const char *tray_name = "Shell_TrayWnd";
			char name[strlen(tray_name) + 1];
			if (GetClassNameA(focus, name, sizeof(name)) &&
				!strcmp(name, tray_name)) {
				focus = NULL;
			}
		}
		focus = focus ? focus : state->focus;
		state->focus = focus;
	}
	DWORD tid = focus ? GetWindowThreadProcessId(focus, NULL) : 0;
	HKL layout = tid ? GetKeyboardLayout(tid) : NULL;
	LCID lcid = layout ? MAKELCID((int) layout & 0xffff, SORT_DEFAULT) : 0;
	lcid = lcid ? lcid : state->lcid;
	bool changed = state->lcid != lcid;
	if (changed) {
		state->lcid = lcid;
		GetLocaleInfoA(lcid, LOCALE_SLOCALIZEDDISPLAYNAME,
			state->notify.szTip, sizeof(state->notify.szTip));
	}
	bool add = state->notify_add;
	if (changed || add) {
		state->notify_add = false;
		state->notify.hIcon = create_icon(&state->icon, lcid);
		Shell_NotifyIconA(add ? NIM_ADD : NIM_MODIFY, &state->notify);
		if (state->notify.hIcon) {
			DestroyIcon(state->notify.hIcon);
			state->notify.hIcon = NULL;
		}
	}
}

static void set_layout(HWND window, HKL layout) {
	SendMessageA(window, WM_INPUTLANGCHANGEREQUEST, 0, (LPARAM) layout);
}

static void set_layout_checked(HWND window, DWORD tid,
	HKL old_layout, HKL layout) {
	set_layout(window, layout);
	HKL new_layout = GetKeyboardLayout(tid);
	if (new_layout && new_layout != layout) {
		set_layout(window, old_layout);
	}
}

static void next_layout(HWND focus) {
	DWORD tid = focus ? GetWindowThreadProcessId(focus, NULL) : 0;
	HKL old_layout = tid ? GetKeyboardLayout(tid) : NULL;
	if (!old_layout) {
		return;
	}
	int count = GetKeyboardLayoutList(0, NULL);
	HKL layouts[count];
	int index = -1;
	if (GetKeyboardLayoutList(count, layouts) == count) {
		for (int i = 0; i < count; i++) {
			if (layouts[i] == old_layout) {
				index = (i + 1) % count;
				break;
			}
		}
	}
	if (index >= 0 && layouts[index] != old_layout) {
		set_layout_checked(focus, tid, old_layout, layouts[index]);
	}
}

static void create_menu(struct state_t *state, HWND window, int x, int y) {
	HMENU menu = CreatePopupMenu();
	if (!menu) {
		return;
	}
	DWORD tid = state->focus
		? GetWindowThreadProcessId(state->focus, NULL) : 0;
	HKL old_layout = tid ? GetKeyboardLayout(tid) : NULL;
	int count = GetKeyboardLayoutList(0, NULL);
	HKL layouts[count];
	if (GetKeyboardLayoutList(count, layouts) == count) {
		char name[80];
		bool inserted = false;
		for (int i = 0; i < count; i++) {
			LCID lcid = MAKELCID((int) layouts[i] & 0xffff,
				SORT_DEFAULT);
			if (!GetLocaleInfoA(lcid, LOCALE_SLOCALIZEDDISPLAYNAME,
				name, sizeof(name))) {
				continue;
			}
			AppendMenuA(menu, layouts[i] == old_layout
				? MF_CHECKED : 0, i + 1, name);
			inserted = true;
		}
		if (inserted && state->exit == EXIT_PASS) {
			AppendMenuA(menu, MF_SEPARATOR, 0, NULL);
		}
	}
	const int id_exit = -1;
	if (state->exit == EXIT_PASS) {
		AppendMenuA(menu, 0, id_exit, "Exit");
	}
	SetForegroundWindow(window);
	int id = TrackPopupMenu(menu, TPM_BOTTOMALIGN | TPM_RIGHTALIGN |
		TPM_NONOTIFY | TPM_RETURNCMD, x, y, 0, window, NULL);
	DestroyMenu(menu);
	if (id > 0 && old_layout && layouts[id - 1] != old_layout) {
		set_layout_checked(state->focus, tid,
			old_layout, layouts[id - 1]);
		update_state(state, window);
	} else if (id == id_exit) {
		state->exit = EXIT_CONFIRM;
	}
}

static LRESULT CALLBACK process_message(HWND window,
	UINT msg, WPARAM wparam, LPARAM lparam) {
	struct state_t *state = (void *) GetWindowLongPtrA(window,
		GWLP_USERDATA);
	static UINT wm_taskbar_created = 0;
	if (!wm_taskbar_created) {
		wm_taskbar_created = RegisterWindowMessageA("TaskbarCreated");
	}
	if (msg == WM_TIMER && wparam == TIMER_ID) {
		update_state(state, window);
		return 0;
	} else if (msg == WM_USER) {
		if (lparam == WM_LBUTTONUP) {
			next_layout(state->focus);
		} else if (lparam == WM_RBUTTONUP) {
			int xy = GetMessagePos();
			int x = (int16_t) xy & 0xffff;
			int y = (int16_t) (xy >> 16) & 0xffff;
			create_menu(state, window, x, y);
		}
		return 0;
	} else if (msg == wm_taskbar_created) {
		state->notify_add = true;
		update_state(state, window);
		return 0;
	}
	return DefWindowProc(window, msg, wparam, lparam);
}

int main(int argc, char **argv) {
	HINSTANCE instance = GetModuleHandle(NULL);
	WNDCLASSEXA wc = {
		.cbSize = sizeof(wc),
		.lpfnWndProc = process_message,
		.hInstance = instance,
		.lpszClassName = "KbdTray"
	};
	if (!RegisterClassExA(&wc)) {
		error_dialog("Failed to register window class");
		return 1;
	}
	HWND window = CreateWindowExA(0, wc.lpszClassName, "KbdTray",
		0, 0, 0, 0, 0, NULL, NULL, instance, NULL);
	if (!window) {
		error_dialog("Failed to create window");
		return 1;
	}
	bool caps = false;
	bool noexit = false;
	for (int i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "caps")) {
			caps = true;
		} else if (!strcmp(argv[i], "noexit")) {
			noexit = true;
		}
	}
	int icon_size = 16;
	HDC dc = CreateCompatibleDC(NULL);
	SetBkMode(dc, TRANSPARENT);
	HFONT font = CreateFontA(16, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, "Trebuchet MS");
	HBITMAP bitmap = CreateDIBSection(dc,
		(BITMAPINFO *) &(BITMAPINFOHEADER) {
			.biSize = sizeof(BITMAPINFOHEADER),
			.biWidth = icon_size,
			.biHeight = icon_size,
			.biPlanes = 1,
			.biBitCount = 32,
			.biCompression = BI_RGB
		},
		DIB_RGB_COLORS,
		NULL, NULL, 0);
	HBITMAP mask = CreateBitmap(icon_size, icon_size, 1, 1, NULL);
	struct state_t state = {
		.icon = {
			.dc = dc,
			.font = font,
			.bitmap = bitmap,
			.mask = mask,
			.caps = caps
		},
		.notify = {
			.cbSize = sizeof(state.notify),
			.hWnd = window,
			.uID = 0,
			.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP,
			.szTip = "",
			.uCallbackMessage = WM_USER
		},
		.notify_add = true,
		.exit = noexit ? EXIT_DISABLE : EXIT_PASS,
	};
	SetWindowLongPtrA(window, GWLP_USERDATA, (LONG_PTR) &state);
	update_state(&state, window);
	SetTimer(window, TIMER_ID, TIMER_INTERVAL, NULL);
	MSG msg;
	while (state.exit != EXIT_CONFIRM && GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	Shell_NotifyIconA(NIM_DELETE, &state.notify);
	KillTimer(window, TIMER_ID);
	DestroyWindow(window);
	DeleteDC(dc);
	DeleteObject(font);
	DeleteObject(bitmap);
	DeleteObject(mask);
	return 0;
}
